# Copyright (c) 2015 Seagate Technology

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

#@author: Ignacio Corderi

import unittest
from kineticpool.core import DeviceInfo
 
class DeviceInfoTests(unittest.TestCase): 
	
	def test_json_decoding_full(self):
		data = '{ "wwn": "123abc", "addresses": [ "192.168.0.121", "192.168.1.121" ], "port": 9999 }'
		info = DeviceInfo.from_json(data)
		
		self.assertEqual(info.wwn, "123abc")
		self.assertEqual(info.addresses, ["192.168.0.121", "192.168.1.121"])
		self.assertEqual(info.port, 9999)
		
	def test_json_decoding_basic(self):
		data = '{ "wwn": "123abc", "addresses": [ "192.168.0.121" ] }'
		info = DeviceInfo.from_json(data)
		
		self.assertEqual(info.wwn, "123abc")
		self.assertEqual(info.addresses, ["192.168.0.121"])
		self.assertEqual(info.port, 8123)		
		
	def test_json_decoding_addresses_empty(self):
		data = '{ "wwn": "123abc", "addresses": [] }'
		with self.assertRaises(ValueError):
			DeviceInfo.from_json(data)	
			
	def test_json_decoding_addresses_missing(self):
		data = '{ "wwn": "123abc" }'
		with self.assertRaises(ValueError):
			DeviceInfo.from_json(data)	
			
	def test_json_decoding_wwn_missing(self):
		data = '{ "addresses": [ "192.168.0.121" ] }'
		with self.assertRaises(ValueError):
			DeviceInfo.from_json(data)								
		
	def test_json_encoding_full(self):
		x = DeviceInfo("123abc",  ["192.168.0.121", "192.168.1.121"], 9999)
		data = x.to_json()
		
		info = DeviceInfo.from_json(data)
		
		self.assertEqual(info.wwn, "123abc")
		self.assertEqual(info.addresses, ["192.168.0.121", "192.168.1.121"])
		self.assertEqual(info.port, 9999)
	
	def test_json_encoding_basic(self):
		x = DeviceInfo("123abc",  ["192.168.0.121"])
		data = x.to_json()
		
		info = DeviceInfo.from_json(data)
		
		self.assertEqual(info.wwn, "123abc")
		self.assertEqual(info.addresses, ["192.168.0.121"])
		self.assertEqual(info.port, 8123)		


# Copyright (c) 2015 Seagate Technology

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

#@author: Ignacio Corderi

import unittest
from kineticpool.core import DeviceInfo
from kineticpool.maps import MemcachedDeviceMap
from kineticpool import exceptions

class MemcachedDeviceMapTests(unittest.TestCase): 

	def test_set_entry(self):
		info = DeviceInfo(wwn="123abc", addresses=["127.0.0.1"])
		m = MemcachedDeviceMap()				
		m["123abc"] = info		
	
	def test_verify_entry(self):
		info = DeviceInfo(wwn="123abc", addresses=["127.0.0.1"])
		m = MemcachedDeviceMap()				
		m["123abc"] = info		
		info2 = m["123abc"]
		
		self.assertEqual(info.wwn, info2.wwn)
		self.assertEqual(info.addresses, info2.addresses)
		self.assertEqual(info.port, info2.port)
		
	def test_invalid_wwn(self):
		info = DeviceInfo(wwn="123abc", addresses=["127.0.0.1"])
		m = MemcachedDeviceMap()				
		m["foo"] = info		
		
		with self.assertRaises(exceptions.InvalidEntry):
			m["foo"]
		
				
	
		
		


import kinetic
import kineticpool
import eventlet
import time
import threading
import socket
from kinetic import AsyncClient
import logging
import struct
import json
from kineticpool.maps import MemcachedDeviceMap
from kineticpool.core import DeviceInfo
import sys 

eventlet.monkey_patch()

class DiscoveredDrive(DeviceInfo):

    def __init__(self, wwn, serial_number, manufacturer, firmware_version, protocol_version, port, addresses):
        super(DiscoveredDrive, self).__init__(wwn, addresses, port)
        self.serial_number = serial_number
        self.manufacturer = manufacturer
        self.firmware_version = firmware_version
        self.protocol_version = protocol_version
        
    @staticmethod
    def from_json(raw): 
        data = json.loads(raw)
        return DiscoveredDrive(data['world_wide_name'],
                               data['serial_number'], data['manufacturer'],
                               data['firmware_version'], data['protocol_version'],
                               data['port'],
                               [str(iface['ipv4_addr']) for iface in data['network_interfaces']])            

    def __str__(self):
        return 'WWN={5}, SN={0}, Version={1}, Proto={2}, Port={3}, Addresses={4}'.format(
                 self.serial_number, self.firmware_version, self.protocol_version,
                 self.port, self.addresses, self.wwn)

class MemcachedHandler(object):
    
    def __init__(self): 
       self.device_map = MemcachedDeviceMap()
    
    def detected(self, drive): 
        self.device_map[drive.wwn] = drive
        
    def new_drive(self, drive): pass 
    
class FileHandler(object):
    
    def __init__(self, path): 
        self.fd = open(path, 'w')            
    
    def new_drive(self, drive): 
        self.fd.write("%s\n" % drive.to_json())
        self.fd.flush()  

class KineticDiscoveryManager(object):
    
    MCAST_ADDR = '239.1.2.3'
    MCAST_PORT = 8123
    
    def __init__(self, interface, logger=None): 
        self.interface = interface
        self.drives = {}     
        self.handlers = []   
        if logger == None:
            logger = logging.getLogger(__name__)
        self.logger = logger
    
    def add_handler(self, handler):
        self.handlers.append(handler)
    
    def run(self): 
        s = joinMcast(self.MCAST_ADDR, self.MCAST_PORT, self.interface)
        t = threading.Thread(target=self._listen, args=(s,))
        t.daemon = True
        t.start()
        t.join()
        s.close()
             
    def _listen(self, s):
        while True:
            raw = s.recv(64 * 2 ** 10)            
        
            d = DiscoveredDrive.from_json(raw)
    
            # notify handlers
            for h in self.handlers:
                if hasattr(h, "detected"): h.detected(d)
                if not (d.wwn in self.drives): 
                    h.new_drive(d)                 
             
            if not (d.wwn in self.drives):                                      
                self.logger.info(d)                 
            else:
                if self.drives[d.wwn].addresses[0] != d.addresses[0]:
                    self.logger.warn('Address changed!! %s' % d)
                                
            # update record     
            self.drives[d.wwn] = d

def joinMcast(mcast_addr,port,if_ip):
    """
    Returns a live multicast socket
    mcast_addr is a dotted string format of the multicast group
    port is an integer of the UDP port you want to receive
    if_ip is a dotted string format of the interface you will use
    """

    #create a UDP socket
    mcastsock = socket.socket(socket.AF_INET,socket.SOCK_DGRAM)

    #allow other sockets to bind this port too
    mcastsock.setsockopt(socket.SOL_SOCKET,socket.SO_REUSEADDR,1)

    #explicitly join the multicast group on the interface specified
    mcastsock.setsockopt(socket.SOL_IP,socket.IP_ADD_MEMBERSHIP,
                socket.inet_aton(mcast_addr)+socket.inet_aton(if_ip))

    #finally bind the socket to start getting data into your socket
    mcastsock.bind((mcast_addr,port))

    return mcastsock

def main():
    import argparse

    parser = argparse.ArgumentParser(description='Kinetic Multicast Discovery Tool')
    parser.add_argument('interface', metavar='interface',
                       help='Address of interface to listen on (i.e. 192.168.0.1)')
    parser.add_argument('--drives', dest='drives', default=None,
                       help='Output discovered devices to a file')
    parser.add_argument('--log', dest='loglevel', default="info",
                       help='Logging level (default=warning)')
    parser.add_argument('--version', action='version', version='%(prog)s ' + kineticpool.__version__)
    
    args = parser.parse_args()

    numeric_level = getattr(logging, args.loglevel.upper(), None)
    if not isinstance(numeric_level, int):
        raise ValueError('Invalid log level: %s' % loglevel)
    logging.basicConfig(format='%(asctime)-8s %(levelname)s: %(message)s', 
                        datefmt="%H:%M:%S", 
                        level=logging.INFO)

    LOG = logging.getLogger(__name__)
   
    mgr = KineticDiscoveryManager(args.interface, logger=LOG)
    mgr.add_handler(MemcachedHandler())
    if args.drives != None:
        mgr.add_handler(FileHandler(args.drives))
    mgr.run()

if __name__ == '__main__':
    main()



# Copyright (c) 2015 Seagate Technology

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

#@author: Ignacio Corderi

import unittest
from kineticpool.core import DeviceInfo
from kineticpool import exceptions
from kineticpool import ConnectionManager
from kineticpool.maps import MemcachedDeviceMap
from kinetic import Client
import utils 

class ConnectionManagerTests(unittest.TestCase): 
	
	@classmethod
	def setUpClass(cls):
		cls._simulator = utils.Simulator(port=8123)
		cls._simulator.start()

	@classmethod
	def tearDownClass(cls):
		cls._simulator.shutdown()

	def setUp(self):
		c = Client()
		c.connect()
		self.device = c.config.worldWideName
		print self.device
		info = DeviceInfo(wwn=self.device, addresses=["127.0.0.1"])
		m = MemcachedDeviceMap()
		m[self.device] = info

	def test_valid_connection(self):		
		mgr = ConnectionManager()
		c = mgr.get_connection(self.device)
		c.put('hello', 'world')
		
	def test_not_found(self):		
		mgr = ConnectionManager()
		with self.assertRaises(exceptions.DeviceNotFound):
			mgr.get_connection("not_found")
			
	def test_wrong_device(self):		
		info = DeviceInfo(wwn="wrong", addresses=["127.0.0.1"])
		mgr = ConnectionManager()
		mgr.device_map["wrong"] = info	
		with self.assertRaises(exceptions.WrongDeviceConnection):
			mgr.get_connection("wrong")	
			
	def test_device_unavailable(self):
		info = DeviceInfo(wwn="unavailable", addresses=["192.168.0.254"])
		mgr = ConnectionManager(connect_timeout=0.01)
		mgr.device_map["unavailable"] = info				
		with self.assertRaises(exceptions.DeviceNotAvailable):
			mgr.get_connection("unavailable")	
			
	def test_persist_connection(self):		
		mgr = ConnectionManager(persist_connection=True)		
		c1 = mgr.get_connection(self.device)
		c2 = mgr.get_connection(self.device)
		self.assertEqual(c1, c2)
		
	def test_non_persistent_connection(self):		
		mgr = ConnectionManager(persist_connection=False)		
		c1 = mgr.get_connection(self.device)
		c2 = mgr.get_connection(self.device)
		self.assertNotEqual(c1.connection_id, c2.connection_id)		
									
		
		
				
	
		
		


# Copyright (c) 2015 Seagate Technology

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

#@author: Ignacio Corderi

import json

class DeviceInfo(object):
    
    def __init__(self, wwn, addresses, port=8123):
        if addresses == None: raise ValueError("Addresses can't be None")
        if len(addresses) == 0: raise ValueError("Addresses can't be None")
        self.wwn = wwn
        self.addresses = addresses
        self.port = port 
       
    @staticmethod        
    def from_json(s): 
        x = json.loads(s)
        if not "wwn" in x: raise ValueError("Missing wwn information") 
        if not "addresses" in x: raise ValueError("Missing address information") 
        return DeviceInfo(x["wwn"], x["addresses"], int(x.get("port", 8123)))
                 
    def to_json(self):
        return json.dumps(self.__dict__)        


from setuptools import setup, find_packages

with open('requirements.txt', 'r') as f:
    requires = [x.strip() for x in f if x.strip()]

from kineticpool import __version__

setup(
    # overview    
    name = 'kineticpool',
    description = "A python connections manager for Kinetic devices",

    # technical info
    version = __version__,
    packages=find_packages(exclude=['test']),
    #requires = requires,
    install_requires=requires,

    # features
     entry_points = {
        'console_scripts': [ 'kinetic-discovery = kineticpool.cmd:main' ],
    },

    # copyright
    author='Ignacio Corderi',
    license='MIT',

    # more info
    url = 'https://github.com/Seagate/kinetic-pool-py',

    # categorization
    keywords = ('kinetic connection pool storage key/value seagate'),
    classifiers  = [
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 2 :: Only',
        'Topic :: Software Development :: Libraries :: Python Modules',
     ],
)



# Copyright (c) 2015 Seagate Technology

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

#@author: Ignacio Corderi

import memcache 
from exceptions import DeviceNotFound, InvalidEntry
from core import DeviceInfo


class MemcachedDeviceMap(object):
    
    AVAILABLE_PREFIX = 'kinetic/available/'
    MEMCACHE_SERVERS = ['127.0.0.1:11211']       
        
    def __init__(self, 
            prefix = AVAILABLE_PREFIX, 
            servers = MEMCACHE_SERVERS):   
        
        self.memcache_client = memcache.Client(servers, debug=0)
        self.available_prefix = prefix
    
    
    def __getitem__(self, device):
        # Spaces are repalced for '-' when adding the WWN to memcached     
        device = device.replace(" ", "-")     
        
        # Caching this will make normal behavior 1ms or 2 faster but 
        # will increase the cost of detecting a fail drive on all 
        # object-servers
        data = self.memcache_client.get(self.available_prefix + str(device))
        
        if data is None:
            raise DeviceNotFound("No entry for device %s." % device)
        
        info = DeviceInfo.from_json(data)
                    
        # sanitizing for memcached
        device = device.replace("-", " ")        
        wwn = info.wwn.replace("-", " ")
        
        # verify key matches entry 
        if wwn != device: raise InvalidEntry("WWN in entry does not match key.")
        
        return info
        
        
    def __setitem__(self, device, info):
        # Spaces are repalced for '-' when adding the WWN to memcached     
        device = device.replace(" ", "-")                  
                
        self.memcache_client.set(self.available_prefix + str(device), info.to_json())
        
        
    def __delitem__(self, device):
        # Spaces are repalced for '-' when adding the WWN to memcached     
        device = device.replace(" ", "-")     
        
        self.memcache_client.delete(self.available_prefix + str(device))

   


# Copyright (c) 2015 Seagate Technology

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

#@author: Ignacio Corderi

import logging
from eventlet import sleep, Timeout
from kinetic import Client

from exceptions import WrongDeviceConnection, DeviceNotAvailable
from maps import MemcachedDeviceMap

LOG = logging.getLogger(__name__)

class ConnectionManager(object):
	
    def __init__(self, 
            persist_connection = False, 
            connect_timeout = 3, 
            connect_retry = 3, 
            map_obj = None, 
            logger = LOG):
        self.logger = logger
        self.persist_connection = persist_connection
        self.connect_timeout = connect_timeout
        self.connect_retry = connect_retry
        if map_obj == None:
            self.device_map = MemcachedDeviceMap()
        else: self.device_map = map_obj
        self.conn_pool = {}
        		
    def _new_connection(self, device, **kwargs):
        kwargs.setdefault('connect_timeout', self.connect_timeout)
        info = self.device_map[device]
        for i in range(1, self.connect_retry + 1):
            try:
                c = Client(info.addresses[0], info.port,**kwargs)
                c.connect()
                if c.config.worldWideName != info.wwn: 
                    raise WrongDeviceConnection("Drive at %s is %s, expected %s." % 
                        (c, c.config.worldWideName, info.wwn))
                return c                        
            except Timeout:
                self.logger.warning('Drive %s connect timeout #%d (%ds)' % (
                    device, i, self.connect_timeout))
            except WrongDeviceConnection: 
                self.logger.exception('Drive %s has an incorrect WWN' % (device))
                self.faulted_device(device)      
                raise        
            except Exception:
                self.logger.exception('Drive %s connection error #%d' % (
                    device, i))
            if i < self.connect_retry:
                sleep(1)
        msg = 'Unable to connect to drive %s after %s attempts' % (
            device, i)
        self.logger.error(msg)
        self.faulted_device(device)        
        raise DeviceNotAvailable(msg)    

    def get_connection(self, device):
        conn = None
        if self.persist_connection:        
            try:
                conn = self.conn_pool[device]
            except KeyError:
                pass
            if conn and conn.faulted:
                conn.close()
                conn = None
            if not conn:
                conn = self.conn_pool[device] = self._new_connection(device)
        else:
            conn = self._new_connection(device)                        
        return conn              
        
    def faulted_device(self, device): 
        del self.device_map[device]
        if self.persist_connection:
            del self.conn_pool[device]    